# coding=utf-8

from flask import Flask

from guniflask.app.initializer import AppInitializer

__all__ = ['create_app']


def create_app(name):
    app = Flask(name)
    app_initializer = AppInitializer(app)
    app_initializer.init()
    return app



# coding=utf-8

__all__ = ['ASYNC_EXECUTOR', 'ASYNC_ANNOTATION_PROCESSOR',
           'TASK_SCHEDULER', 'SCHEDULED_ANNOTATION_PROCESSOR']

ASYNC_EXECUTOR = '__async_executor'
ASYNC_ANNOTATION_PROCESSOR = '__async_annotation_processor'

TASK_SCHEDULER = '__task_scheduler'
SCHEDULED_ANNOTATION_PROCESSOR = '__scheduled_annotation_processor'



# coding=utf-8

from flask import request, _request_ctx_stack
from werkzeug.local import LocalProxy

from guniflask.security.authentication import Authentication, PreAuthenticatedToken
from guniflask.security.token import OAuth2AccessToken

__all__ = ['current_auth', 'AuthenticationManager', 'OAuth2AuthenticationManager',
           'BearerTokenExtractor']


def _load_authentication():
    ctx = _request_ctx_stack.top
    if ctx is not None:
        if not hasattr(ctx, 'authentication'):
            return None
        return ctx.authentication


current_auth = LocalProxy(_load_authentication)


class AuthenticationManager:
    def authenticate(self, authentication: Authentication) -> Authentication:
        raise NotImplemented


class OAuth2AuthenticationManager(AuthenticationManager):
    def __init__(self):
        self.token_service = None
        self.token_extractor = BearerTokenExtractor()

    def init_app(self, app):
        app.before_request(self.do_authentication_filter)

    def authenticate(self, authentication):
        # TODO
        pass

    def do_authentication_filter(self):
        authentication = self.token_extractor.extract()
        auth_result = self.authenticate(authentication)
        ctx = _request_ctx_stack.top
        if ctx is not None:
            ctx.authentication = auth_result


class BearerTokenExtractor:
    def extract(self):
        token = self._extract_token_from_header()
        if token is None:
            token = self._extract_token_from_query()
        if token is not None:
            return PreAuthenticatedToken(token)

    @staticmethod
    def _extract_token_from_header():
        auth = request.headers.get('Authorization')
        if auth is not None and auth.lower().startswith(OAuth2AccessToken.BEARER_TYPE.lower()):
            return auth.split(' ', 1)[1]

    @staticmethod
    def _extract_token_from_query():
        return request.args.get(OAuth2AccessToken.ACCESS_TOKEN)



# coding=utf-8

import inspect

import sqlalchemy

__all__ = ['DatabaseDialect', 'supported_dialects']


class DatabaseDialect:
    column_types = {}

    def get_kwargs_of_column_type(self, coltype):
        argspec = inspect.getfullargspec(coltype.__class__.__init__)
        defaults = dict(zip(argspec.args[-len(argspec.defaults or ()):],
                            argspec.defaults or ()))
        kwargs = {}
        missing = object()
        for attr in argspec.args[1:]:
            value = getattr(coltype, attr, missing)
            default = defaults.get(attr, missing)
            if value is not missing and value != default:
                kwargs[attr] = value
        return kwargs

    def convert_column_type(self, coltype):
        coltype_name = coltype.__class__.__name__
        if coltype_name not in self.column_types:
            raise ValueError('Unsupported column type: {}'.format(coltype_name))
        column_type = self.column_types[coltype_name]

        args = []

        if issubclass(column_type, sqlalchemy.Enum):
            args.extend(repr(arg) for arg in coltype.enums)
            if coltype.name is not None:
                args.append('name={!r}'.format(coltype.name))
        else:
            # All other types
            kwargs = self.get_kwargs_of_column_type(coltype)
            argspec = inspect.getfullargspec(column_type.__init__)
            defaults = dict(zip(argspec.args[-len(argspec.defaults or ()):],
                                argspec.defaults or ()))
            missing = object()
            use_kwargs = False
            for attr in argspec.args[1:]:
                if attr.startswith('_'):
                    continue
                value = kwargs.get(attr, missing)
                default = defaults.get(attr, missing)
                if value is missing or value == default:
                    use_kwargs = True
                elif use_kwargs:
                    args.append('{}={!r}'.format(attr, value))
                else:
                    args.append(repr(value))
        rendered = column_type.__name__
        if args:
            rendered += '({})'.format(', '.join(args))
        return rendered


class MysqlDialect(DatabaseDialect):
    column_types = {
        'BIGINT': sqlalchemy.BigInteger,
        'BINARY': sqlalchemy.Binary,
        'BIT': sqlalchemy.BOOLEAN,
        'BLOB': sqlalchemy.LargeBinary,
        'BOOLEAN': sqlalchemy.Boolean,
        'CHAR': sqlalchemy.CHAR,
        'DATE': sqlalchemy.Date,
        'DATETIME': sqlalchemy.DateTime,
        'DECIMAL': sqlalchemy.DECIMAL,
        'DOUBLE': sqlalchemy.Float,
        'ENUM': sqlalchemy.Enum,
        'FLOAT': sqlalchemy.Float,
        'INTEGER': sqlalchemy.Integer,
        'JSON': sqlalchemy.JSON,
        'LONGBLOB': sqlalchemy.LargeBinary,
        'LONGTEXT': sqlalchemy.Text,
        'MEDIUMTEXT': sqlalchemy.Text,
        'NUMERIC': sqlalchemy.Numeric,
        'SMALLINT': sqlalchemy.SmallInteger,
        'TEXT': sqlalchemy.Text,
        'TIME': sqlalchemy.TIME,
        'TIMESTAMP': sqlalchemy.TIMESTAMP,
        'TINYBLOB': sqlalchemy.Binary,
        'TINYINT': sqlalchemy.SmallInteger,
        'VARBINARY': sqlalchemy.VARBINARY,
        'VARCHAR': sqlalchemy.String,
    }


supported_dialects = {
    'mysql': MysqlDialect()
}



# coding=utf-8

import inspect
from typing import Type

from guniflask.annotation.core import Annotation
from guniflask.annotation.annotation_utils import AnnotationUtils
from guniflask.context.condition import Condition

__all__ = ['Bean', 'bean',
           'Component', 'component',
           'Configuration', 'configuration',
           'Conditional', 'conditional',
           'Repository', 'repository',
           'Service', 'service',
           'Controller', 'controller']


class Bean(Annotation):
    def __init__(self, name=None):
        super().__init__(name=name)


def bean(name: str = None):
    def wrap_func(func):
        AnnotationUtils.add_annotation(func, Bean(name=name))
        return func

    if inspect.isclass(name) or inspect.isfunction(name):
        f = name
        name = None
        return wrap_func(f)
    return wrap_func


class Component(Annotation):
    def __init__(self, name=None):
        super().__init__(name=name)


def component(name: str = None):
    def wrap_func(func):
        AnnotationUtils.add_annotation(func, Component(name=name))
        return func

    if inspect.isclass(name) or inspect.isfunction(name):
        f = name
        name = None
        return wrap_func(f)
    return wrap_func


class Configuration(Component):
    def __init__(self, name=None):
        super().__init__(name=name)


def configuration(name: str = None):
    def wrap_func(func):
        AnnotationUtils.add_annotation(func, Configuration(name=name))
        return func

    if inspect.isclass(name) or inspect.isfunction(name):
        f = name
        name = None
        return wrap_func(f)
    return wrap_func


class Conditional(Annotation):
    def __init__(self, condition: Type[Condition]):
        super().__init__(condition=condition)


def conditional(condition: Type[Condition]):
    def wrap_func(func):
        AnnotationUtils.add_annotation(func, Conditional(condition))
        return func

    return wrap_func


class Repository(Component):
    def __init__(self, name=None):
        super().__init__(name=name)


def repository(name: str = None):
    def wrap_func(func):
        AnnotationUtils.add_annotation(func, Repository(name=name))
        return func

    if inspect.isclass(name) or inspect.isfunction(name):
        f = name
        name = None
        return wrap_func(f)
    return wrap_func


class Service(Component):
    def __init__(self, name=None):
        super().__init__(name=name)


def service(name: str = None):
    def wrap_func(func):
        AnnotationUtils.add_annotation(func, Service(name=name))
        return func

    if inspect.isclass(name) or inspect.isfunction(name):
        f = name
        name = None
        return wrap_func(f)
    return wrap_func


class Controller(Component):
    def __init__(self, name=None):
        super().__init__(name=name)


def controller(name: str = None):
    def wrap_func(func):
        AnnotationUtils.add_annotation(func, Controller(name=name))
        return func

    if inspect.isclass(name) or inspect.isfunction(name):
        f = name
        name = None
        return wrap_func(f)
    return wrap_func



# coding=utf-8

import os
import logging
from os.path import isfile, join

from flask import current_app
from werkzeug.local import LocalProxy

from guniflask.config.setting import Settings

log = logging.getLogger(__name__)

__all__ = ['settings', 'SETTINGS', 'app_default_settings', 'AppConfig', 'load_config', 'load_profile_config']

settings = LocalProxy(lambda: current_app.extensions['settings'])


class SETTINGS:
    DEBUG = 'debug'
    CORS = 'cors'
    JWT = 'jwt'
    WRAP_SQLALCHEMY_MODEL = 'wrap_sqlalchemy_model'


app_default_settings = {
    SETTINGS.DEBUG: False,
    SETTINGS.CORS: True,
    SETTINGS.JWT: False,
    SETTINGS.WRAP_SQLALCHEMY_MODEL: True,
    # Flask-SQLAlchemy
    'SQLALCHEMY_TRACK_MODIFICATIONS': False
}


class AppConfig:
    def __init__(self, app=None):
        if app:
            self.init_app(app)

    def init_app(self, app):
        s = Settings(self._load_app_settings(app.name))
        self._set_default_settings(s)
        app.extensions['settings'] = s

    @property
    def settings(self):
        return self.app_settings(current_app)

    def app_settings(self, app):
        return app.extensions['settings']

    def _load_app_settings(self, app_name):
        c = {}
        conf_dir = os.environ.get('GUNIFLASK_CONF_DIR')
        active_profiles = os.environ.get('GUNIFLASK_ACTIVE_PROFILES')
        kwargs = self._get_default_settings_from_env()
        if conf_dir:
            c = load_profile_config(conf_dir, app_name, profiles=active_profiles, **kwargs)
            c['active_profiles'] = active_profiles
        c.update(kwargs)
        s = {}
        for name in c:
            if not name.startswith('_'):
                s[name] = c[name]
        return s

    def _get_default_settings_from_env(self):
        kwargs = {'home': os.environ.get('GUNIFLASK_HOME', os.curdir),
                  'project_name': os.environ.get('GUNIFLASK_PROJECT_NAME')}
        if os.environ.get('GUNIFLASK_DEBUG'):
            kwargs['debug'] = True
        else:
            kwargs['debug'] = False
        return kwargs

    def _set_default_settings(self, s):
        for k, v in app_default_settings.items():
            s.setdefault(k, v)


def load_config(fname, **kwargs):
    if fname is None or not isfile(fname):
        raise FileNotFoundError("Cannot find configuration file '{}'".format(fname))
    code = compile(open(fname, 'rb').read(), fname, 'exec')
    cfg = {
        "__builtins__": __builtins__,
        "__name__": "__config__",
        "__file__": fname,
        "__doc__": None,
        "__package__": None
    }
    cfg.update(kwargs)
    exec(code, cfg, cfg)
    return cfg


def load_profile_config(conf_dir, name, profiles=None, **kwargs):
    pc = load_config(join(conf_dir, name + '.py'), **kwargs)
    if profiles:
        profiles = profiles.split(',')
        profiles.reverse()
        for profile in profiles:
            if profile:
                pc_file = join(conf_dir, name + '_' + profile + '.py')
                if isfile(pc_file):
                    c = load_config(pc_file, **kwargs)
                    pc.update(c)
    return pc



# coding=utf-8

from .definition import *
from .definition_registry import *
from .errors import *
from .factory import *
from .factory_hook import *
from .factory_post_processor import *
from .name_generator import *
from .post_processor import *
from .registry import *

__all__ = (definition.__all__ +
           definition_registry.__all__ +
           errors.__all__ +
           factory.__all__ +
           factory_hook.__all__ +
           factory_post_processor.__all__ +
           name_generator.__all__ +
           post_processor.__all__ +
           registry.__all__)



# coding=utf-8

import os
import signal

from guniflask.utils.cli import readchar

__all__ = ['Step', 'StepChain', 'InputStep', 'ChoiceStep']


class Step:
    desc = None

    def __init__(self):
        self.value = None
        self.title = self.desc
        self.tooltip = None

    def process_arguments(self, settings):
        pass

    def process_user_input(self):
        while True:
            self.show_question()
            user_input = self.get_user_input()
            if self.check_user_input(user_input):
                self.show_decision()
                break
            else:
                self.show_invalid_tooltip()

    def show_question(self):
        raise NotImplementedError

    def get_user_input(self):
        raise NotImplementedError

    def show_decision(self):
        raise NotImplementedError

    def show_invalid_tooltip(self):
        raise NotImplementedError

    def check_user_input(self, user_input):
        return True

    def update_settings(self, settings):
        raise NotImplementedError

    def question(self):
        return '\033[32m?\033[0m {} \033[37m({})\033[0m'.format(self.title, self.tooltip)

    def decision(self):
        return '\033[32m?\033[0m {} \033[36m{}\033[0m'.format(self.title, self.value)

    def go_back_lines(self, n=1):
        return '\033[1A\r\033[K' * n

    def clean_line(self):
        return '\r\033[K'

    @property
    def next_step(self):
        return None


class StepChain:
    def __init__(self, step, length=0, next_steps=None):
        self.step = step
        self.length = length
        self.next_steps = next_steps

    def previous(self, father, siblings=None):
        max_depth = self.length
        next_steps = [self]
        if siblings:
            for s in siblings:
                max_depth = max(max_depth, s.length)
                next_steps.append(s)
        s = StepChain(father, length=max_depth + 1, next_steps=next_steps)
        return s

    def __iter__(self):
        return StepChainIter(self)


class StepChainIter:
    def __init__(self, step_chain):
        self.step_chain = step_chain
        self.total_steps = step_chain.length + 1

    def __next__(self):
        if not self.step_chain:
            raise StopIteration
        cur_step = self.total_steps - self.step_chain.length
        step = self.step_chain.step

        next_step = None
        if self.step_chain.next_steps:
            if len(self.step_chain.next_steps) == 1:
                next_step = self.step_chain.next_steps[0]
            else:
                next_step_cls = self.step_chain.step.next_step
                if next_step_cls:
                    for s in self.step_chain.next_steps:
                        if isinstance(s.step, next_step_cls):
                            next_step = s
                            break
                if next_step is None:
                    raise RuntimeError('No matching next step')
        self.step_chain = next_step

        return cur_step, self.total_steps, step


class InputStep(Step):
    def show_question(self):
        print(self.question() + ' ', end='', flush=True)

    def get_user_input(self):
        return input()

    def show_decision(self):
        print(self.clean_line(), end='', flush=True)
        print(self.go_back_lines(), end='', flush=True)
        print(self.decision(), flush=True)

    def show_invalid_tooltip(self):
        print(self.go_back_lines(), end='', flush=True)


class ChoiceStep(Step):
    def __init__(self):
        super().__init__()
        self.selected = 0
        self.choices = []
        self.values = []

    def show_question(self):
        print(self.question() + ' ', end='', flush=True)
        for i in range(len(self.choices)):
            print(flush=True)
            if i == self.selected:
                print('\033[36m>\033[0m \033[36m{}\033[0m'.format(self.choices[i]), end='', flush=True)
            else:
                print('  {}'.format(self.choices[i]), end='', flush=True)
        self.hide_cursor()

    def get_user_input(self):
        ch = readchar()
        if ch == '\x03':
            self.show_cursor()
            raise KeyboardInterrupt
        if ch == '\x1a':
            self.show_cursor()
            os.kill(os.getpid(), signal.SIGSTOP)
        return ch

    def check_user_input(self, user_input):
        if user_input == '\r' or user_input == '\n':
            self.value = self.choices[self.selected]
            self.show_cursor()
            return True
        if user_input == '\x1b':
            assert readchar() == '['
            ch = readchar()
            if ch == 'A':
                self.selected -= 1
                if self.selected < 0:
                    self.selected = len(self.choices) - 1
            elif ch == 'B':
                self.selected += 1
                if self.selected >= len(self.choices):
                    self.selected = 0
        return False

    def show_decision(self):
        print(self.clean_line(), end='', flush=True)
        print(self.go_back_lines(len(self.choices)), end='', flush=True)
        print(self.decision(), flush=True)

    def show_invalid_tooltip(self):
        print(self.clean_line(), end='', flush=True)
        print(self.go_back_lines(len(self.choices)), end='', flush=True)

    def add_choice(self, desc=None, value=None):
        self.choices.append(desc)
        self.values.append(value)

    @property
    def selected_value(self):
        return self.values[self.selected]

    def show_cursor(self):
        print('\033[?25h', end='', flush=True)

    def hide_cursor(self):
        print('\033[?25l', end='', flush=True)



# coding=utf-8

from .bind_annotation import *
from .blueprint_post_processor import *
from .context import *
from .global_singleton_annotation import *
from .scheduling_config import *

__all__ = (bind_annotation.__all__ +
           blueprint_post_processor.__all__ +
           context.__all__ +
           global_singleton_annotation.__all__ +
           scheduling_config.__all__)



# coding=utf-8

from flask import current_app, _request_ctx_stack
from werkzeug.local import LocalProxy

from guniflask.security.authentication import OAuth2Authentication, UserAuthentication
from guniflask.security.authentication_manager import current_auth

__all__ = ['current_user', 'User']


def _load_user():
    ctx = _request_ctx_stack.top
    if ctx is not None:
        if not hasattr(ctx, 'user'):
            user = AnonymousUser()
            auth = current_auth._get_current_object()
            if auth is not None:
                if isinstance(auth, OAuth2Authentication):
                    auth = auth.user_authentication
                if isinstance(auth, UserAuthentication):
                    user = User(username=auth.principal, authorities=auth.authorities,
                                details=auth.details)
            ctx.user = user
        return ctx.user


current_user = LocalProxy(_load_user)


class User:
    user_role_prefix = 'role_'

    def __init__(self, username=None, authorities=None, details=None):
        self._username = username
        self._authorities = set()
        if authorities:
            self._authorities.update(authorities)
        self.details = details

    @property
    def username(self):
        return self._username

    @property
    def authorities(self):
        return self._authorities

    def has_authority(self, authority):
        return self.has_any_authority(authority)

    def has_any_authority(self, *authorities):
        return self._has_any_authority_name(None, *authorities)

    def has_role(self, role):
        return self.has_any_role(role)

    def has_any_role(self, *roles):
        prefix = current_app.config.get('USER_ROLE_PREFIX', self.user_role_prefix)
        return self._has_any_authority_name(prefix, *roles)

    def _has_any_authority_name(self, prefix, *roles):
        for role in roles:
            r = self._role_with_prefix(prefix, role)
            if r in self.authorities:
                return True
        return False

    @staticmethod
    def _role_with_prefix(prefix, role):
        if role is None or prefix is None:
            return role
        if role.startswith(prefix):
            return role
        return prefix + role


class AnonymousUser(User):
    pass



# coding=utf-8

from .access_annotation import *
from .authentication import *
from .authentication_manager import *
from .client_details import *
from .errors import *
from .jwt import *
from .request import *
from .request_factory import *
from .token import *
from .token_converter import *
from .token_service import *
from .token_store import *
from .user_details import *

__all__ = (access_annotation.__all__ +
           authentication.__all__ +
           authentication_manager.__all__ +
           client_details.__all__ +
           errors.__all__ +
           jwt.__all__ +
           request.__all__ +
           request_factory.__all__ +
           token.__all__ +
           token_converter.__all__ +
           token_service.__all__ +
           token_store.__all__ +
           user_details.__all__)



# coding=utf-8

import os


def daemonize():
    if os.fork():
        os._exit(0)
    os.setsid()
    if os.fork():
        os._exit(0)
    os.umask(0o22)
    os.closerange(0, 3)
    fd_null = os.open(os.devnull, os.O_RDWR)
    if fd_null != 0:
        os.dup2(fd_null, 0)
    os.dup2(fd_null, 1)
    os.dup2(fd_null, 2)


def pid_exists(pid):
    try:
        os.kill(pid, 0)
    except OSError:
        return False
    return True



# coding=utf-8

from typing import Mapping

from guniflask.security.authentication import Authentication, OAuth2Authentication
from guniflask.security.token import OAuth2AccessToken

__all__ = ['AccessTokenConverter', 'TokenEnhancer', 'UserAuthenticationConverter',
           'JwtAccessTokenConverter']


class AccessTokenConverter:
    """
    Converter interface for token service implementations that store authentication data inside the token.
    """

    AUD = 'aud'
    CLIENT_ID = 'client_id'
    EXP = 'exp'
    JTI = 'jti'
    ATI = 'ati'
    SCOPE = OAuth2AccessToken.SCOPE
    AUTHORITIES = 'authorities'

    def convert_access_token(self, access_token: OAuth2AccessToken, authentication: OAuth2Authentication) -> Mapping:
        # TODO
        pass

    def extract_access_token(self, token_value: str, data: Mapping) -> OAuth2AccessToken:
        # TODO
        pass

    def extract_authentication(self, data: Mapping) -> OAuth2Authentication:
        # TODO
        pass


class TokenEnhancer:
    def enhance(self, access_token: OAuth2AccessToken, authentication: OAuth2Authentication) -> OAuth2AccessToken:
        raise NotImplemented


class UserAuthenticationConverter:
    """
    Utility interface for converting a user authentication to and from a Map.
    """

    USERNAME = 'username'

    def convert_user_authentication(self, user_authentication: Authentication):
        # TODO
        pass

    def extract_authentication(self, data):
        # TODO
        pass


class JwtAccessTokenConverter(AccessTokenConverter, TokenEnhancer):
    TOKEN_ID = AccessTokenConverter.JTI
    ACCESS_TOKEN_ID = AccessTokenConverter.ATI

    def convert_access_token(self, access_token, authentication):
        # TODO
        pass

    def extract_access_token(self, token_value, data):
        # TODO
        pass

    def extract_authentication(self, data):
        # TODO
        pass

    def _enhance(self, access_token, authentication):
        # TODO
        pass

    def _encode(self, access_token, authentication):
        # TODO
        pass

    def _decode(self, token_value):
        # TODO
        pass

    def is_refresh_token(self, token):
        return self.ACCESS_TOKEN_ID in token.additional_information()



# coding=utf-8

__all__ = ['OAuth2Request']


class BaseRequest:
    def __init__(self, client_id=None, scope=None, request_parameters=None):
        self._client_id = client_id
        self._scope = set()
        if scope is not None:
            self._scope.update(scope)
        self._request_parameters = {}
        if request_parameters is not None:
            self._request_parameters.update(request_parameters)

    @property
    def client_id(self):
        return self._client_id

    @property
    def scope(self):
        return self._scope

    @property
    def request_parameters(self):
        return self._request_parameters


class OAuth2Request(BaseRequest):
    def __init__(self, authorities=None, approved=None, resource_ids=None, redirect_uri=None, response_types=None,
                 **kwargs):
        super().__init__(**kwargs)
        self._authorities = set()
        if authorities is not None:
            self._authorities.update(authorities)
        self._approved = False
        if approved is not None:
            self._approved = approved
        self._resource_ids = {}
        if resource_ids is not None:
            self._resource_ids.update(resource_ids)
        self._redirect_uri = redirect_uri
        self._response_types = set()
        if response_types is not None:
            self._response_types.update(response_types)
        self._refresh = None

    @property
    def is_approved(self):
        return self._approved

    @property
    def authorities(self):
        return self._authorities

    @property
    def resource_ids(self):
        return self._resource_ids

    @property
    def redirect_uri(self):
        return self._redirect_uri

    @property
    def response_types(self):
        return self._response_types



# coding=utf-8

import os
import base64
import datetime as dt
import uuid

import jwt
from flask import current_app, _request_ctx_stack
from werkzeug.local import LocalProxy

from guniflask.security.authentication import UserAuthentication
from guniflask.security.authentication_manager import AuthenticationManager, BearerTokenExtractor
from guniflask.security.token import OAuth2AccessToken
from guniflask.security.errors import InvalidTokenError
from guniflask.security.token_converter import AccessTokenConverter, JwtAccessTokenConverter, \
    UserAuthenticationConverter

__all__ = ['JwtHelper', 'jwt_manager', 'JwtManager']


class JwtHelper:
    @staticmethod
    def generate_jwt_secret(n=32):
        return base64.b64encode(os.urandom(n), altchars=b'-_').decode().replace('=', '')

    @staticmethod
    def decode_jwt(token, key, algorithm, **kwargs):
        return jwt.decode(token, key=key, algorithms=[algorithm], **kwargs)

    @staticmethod
    def encode_jwt(payload, key, algorithm, **kwargs):
        token = jwt.encode(payload, key, algorithm=algorithm, **kwargs)
        if isinstance(token, bytes):
            token = token.decode('utf-8')
        return token


jwt_manager = LocalProxy(lambda: current_app.extensions['jwt_manager'])


class JwtManager(AuthenticationManager):
    EXP = AccessTokenConverter.EXP
    JTI = AccessTokenConverter.JTI
    AUTHORITIES = AccessTokenConverter.AUTHORITIES
    USERNAME = UserAuthenticationConverter.USERNAME
    USER_DETAILS = 'user_details'

    def __init__(self):
        self.token_extractor = BearerTokenExtractor()
        self.token_converter = JwtAccessTokenConverter()

    def init_app(self, app):
        self._set_default_app_config(app)
        app.before_request(self.do_authentication_filter)
        app.extensions['jwt_manager'] = self

    def _set_default_app_config(self, app):
        app.config.setdefault('JWT_ALGORITHM', 'HS256')
        app.config.setdefault('JWT_SECRET', None)
        app.config.setdefault('JWT_PRIVATE_KEY', None)
        app.config.setdefault('JWT_PUBLIC_KEY', None)
        app.config.setdefault('ACCESS_TOKEN_EXPIRES_IN', dt.timedelta(days=1))
        app.config.setdefault('REFRESH_TOKEN_EXPIRES_IN', dt.timedelta(days=365))

    def create_access_token(self, authorities=None, username=None, user_details=None) -> str:
        app_config = current_app.config

        expires_in = app_config['ACCESS_TOKEN_EXPIRES_IN']
        payload = {
            self.JTI: str(uuid.uuid4())
        }
        exp = dt.datetime.utcnow() + expires_in
        payload[self.EXP] = exp

        payload[self.AUTHORITIES] = authorities
        payload[self.USERNAME] = username
        payload[self.USER_DETAILS] = user_details

        return JwtHelper.encode_jwt(payload, app_config['JWT_SECRET'] or app_config['JWT_PRIVATE_KEY'],
                                    app_config['JWT_ALGORITHM'])

    def read_access_token(self, access_token_value: str) -> OAuth2AccessToken:
        payload = self._decode(access_token_value)
        return self.token_converter.extract_access_token(access_token_value, payload)

    def authenticate(self, authentication):
        access_token_value = authentication.principal
        payload = self._decode(access_token_value)
        authorities = payload[self.AUTHORITIES]
        username = payload[self.USERNAME]
        user_details = payload[self.USER_DETAILS]
        user_auth = UserAuthentication(username, authorities=authorities)
        user_auth.details = user_details
        user_auth.authenticate(True)
        return user_auth

    def _decode(self, access_token_value):
        app_config = current_app.config
        try:
            payload = JwtHelper.decode_jwt(access_token_value, app_config['JWT_SECRET'] or app_config['JWT_PUBLIC_KEY'],
                                           app_config['JWT_ALGORITHM'])
        except Exception as e:
            raise InvalidTokenError(e)
        return payload

    def do_authentication_filter(self):
        auth = self.token_extractor.extract()
        if auth is not None:
            user_auth = self.authenticate(auth)
            ctx = _request_ctx_stack.top
            if ctx is not None:
                ctx.authentication = user_auth



# coding=utf-8

import sys
import os
from os.path import isfile, join
import argparse
import signal
import time

from sqlalchemy.schema import MetaData

from guniflask.utils.env import walk_modules
from guniflask.model.sqlgen import SqlToModelGenerator
from guniflask.cli.errors import UsageError
from guniflask.cli.command import Command
from guniflask.app import create_app
from guniflask.gunicorn import GunicornApplication
from guniflask.utils.process import pid_exists
from guniflask.cli.env import set_default_env, get_project_name_from_env


class InitDb(Command):
    @property
    def name(self):
        return 'initdb'

    @property
    def syntax(self):
        return '[options]'

    @property
    def short_desc(self):
        return 'Initialize database from definition of models'

    def add_arguments(self, parser):
        parser.add_argument('-p', '--active-profiles', dest='active_profiles', metavar='PROFILES',
                            help='active profiles (comma-separated)')
        parser.add_argument('-b', '--bind', dest='bind', metavar='KEY',
                            help='bind key')
        parser.add_argument('-f', '--force', dest='force', action='store_true', default=False,
                            help='force creating all tables')

    def process_arguments(self, args):
        if args.active_profiles:
            os.environ['GUNIFLASK_ACTIVE_PROFILES'] = args.active_profiles
        os.environ.setdefault('GUNIFLASK_ACTIVE_PROFILES', 'dev')

    def run(self, args):
        project_name = get_project_name_from_env()

        walk_modules(project_name)
        app = create_app(project_name)
        with app.app_context():
            s = app.extensions.get('sqlalchemy')
            if not s:
                raise UsageError('Not found sqlalchemy')
            db = s.db
            if args.force:
                db.drop_all()
            else:
                print("\033[33mThe tables already exist will be skipped.\033[0m")
                print("\033[33mYou can try '-f' option to force creating all tables.\033[0m")
            db.create_all(bind=args.bind)


class TableToModel(Command):
    @property
    def name(self):
        return 'table2model'

    @property
    def syntax(self):
        return '[options]'

    @property
    def short_desc(self):
        return 'Convert database tables to definition of models'

    def add_arguments(self, parser):
        parser.add_argument('--tables', dest='tables',
                            help='tables to process (comma-separated, default: all)')
        parser.add_argument('--dest', dest='dest',
                            help='where to put the models generated from database tables')
        parser.add_argument('-p', '--active-profiles', dest='active_profiles', metavar='PROFILES',
                            help='active profiles (comma-separated)')
        parser.add_argument('-b', '--bind', dest='bind', metavar='KEY',
                            help='bind key')

    def process_arguments(self, args):
        if args.tables is not None:
            args.tables = args.tables.split(',')
        if args.active_profiles:
            os.environ['GUNIFLASK_ACTIVE_PROFILES'] = args.active_profiles
        os.environ.setdefault('GUNIFLASK_ACTIVE_PROFILES', 'dev')

    def run(self, args):
        project_name = get_project_name_from_env()

        app = create_app(project_name)
        with app.app_context():
            settings = app.extensions['settings']
            s = app.extensions.get('sqlalchemy')
            if not s:
                raise UsageError('Not found sqlalchemy')
            db = s.db

            dest = []
            default_dest = join(project_name, 'models')
            if args.tables:
                config_dest = args.dest
                if config_dest is None:
                    config_dest = settings.get('table2model_dest')
                if not isinstance(config_dest, str):
                    config_dest = default_dest
                dest.append({'tables': args.tables, 'dest': config_dest, 'bind': args.bind})
            elif args.dest:
                dest.append({'tables': None, 'dest': args.dest, 'bind': args.bind})
            else:
                config_dest = settings.get('table2model_dest', default_dest)
                if isinstance(config_dest, str):
                    dest.append({'tables': None, 'dest': config_dest})
                else:
                    for d in config_dest:
                        t = d.get('tables')
                        if isinstance(t, str):
                            t = t.split(',')
                        dest.append({'tables': t, 'dest': d.get('dest', default_dest),
                                     'bind': d.get('bind')})

            for d in dest:
                engine = db.get_engine(bind=d.get('bind'))
                metadata = MetaData(engine)
                metadata.reflect(only=d.get('tables'))
                gen = SqlToModelGenerator(project_name, metadata, bind=d.get('bind'))
                gen.render(join(settings['home'], d.get('dest')))


class Debug(Command):
    @property
    def name(self):
        return 'debug'

    @property
    def short_desc(self):
        return 'Debug application'

    def add_arguments(self, parser):
        parser.add_argument('-d', '--daemon', dest='daemon', action='store_true', help='run in daemon mode')
        parser.add_argument('-p', '--active-profiles', dest='active_profiles', metavar='PROFILES',
                            help='active profiles (comma-separated)')

    def process_arguments(self, args):
        if args.active_profiles:
            os.environ['GUNIFLASK_ACTIVE_PROFILES'] = args.active_profiles
        os.environ['GUNIFLASK_DEBUG'] = '1'
        os.environ.setdefault('GUNIFLASK_ACTIVE_PROFILES', 'dev')

    def run(self, args):
        app = GunicornApplication()
        if args.daemon:
            app.set_option('daemon', True)
        pid = _read_pid(app.options)
        if pid is not None and pid_exists(pid):
            print('Application is already started')
            self.exitcode = 1
        else:
            app.run()


class Start(Command):
    @property
    def name(self):
        return 'start'

    @property
    def short_desc(self):
        return 'Start application'

    def add_arguments(self, parser):
        parser.add_argument('--daemon-off', dest='daemon_off', action='store_true', help='turn off daemon mode')
        parser.add_argument('-p', '--active-profiles', dest='active_profiles', metavar='PROFILES',
                            help='active profiles (comma-separated)')

    def process_arguments(self, args):
        if args.active_profiles:
            os.environ['GUNIFLASK_ACTIVE_PROFILES'] = args.active_profiles
        os.environ.setdefault('GUNIFLASK_ACTIVE_PROFILES', 'prod')

    def run(self, args):
        app = GunicornApplication()
        if args.daemon_off:
            app.set_option('daemon', False)
        pid = _read_pid(app.options)
        if pid is not None and pid_exists(pid):
            print('Application is already started')
            self.exitcode = 1
        else:
            app.run()


class Stop(Command):
    @property
    def name(self):
        return 'stop'

    @property
    def short_desc(self):
        return 'Stop application'

    def run(self, args):
        app = GunicornApplication()
        pid = _read_pid(app.options)
        if pid is None or not pid_exists(pid):
            print('No application to stop')
            self.exitcode = 1
        else:
            print('kill {}'.format(pid))
            os.kill(pid, signal.SIGTERM)
            time.sleep(3)
            try:
                os.kill(pid, 0)
            except OSError:
                pass
            else:
                print('Application did not stop gracefully after 3 seconds')
                print('kill -9 {}'.format(pid))
                os.kill(pid, signal.SIGKILL)


def _read_pid(options):
    pidfile = options.get('pidfile')
    if isfile(pidfile):
        with open(pidfile, 'r') as f:
            line = f.readline()
            if line:
                pid = line.strip()
                if pid:
                    return int(pid)


def _get_commands_from_module():
    d = {}
    for cmd in globals().values():
        if isinstance(cmd, type) and issubclass(cmd, Command):
            o = cmd()
            if o.name:
                d[o.name] = o
    return d


def _print_commands():
    print("Usage: guniflask-manage <command> [options] [args]\n")
    print("Available commands:")
    cmds = _get_commands_from_module()
    for cmdname, cmdclass in sorted(cmds.items()):
        print("   {:<13} {}".format(cmdname, cmdclass.short_desc))
    print()
    print('Use "guniflask-manage <command> -h" to see more info about a command')


def _print_unknown_command(cmdname):
    print("Unknown command: %s\n" % cmdname, file=sys.stderr)
    print('Use "guniflask-manage" to see available commands', file=sys.stderr)


def main(argv=None):
    set_default_env()
    if argv is None:
        argv = sys.argv
    cmds = _get_commands_from_module()
    cmdname = argv[1] if len(argv) > 1 else None
    if not cmdname or cmdname in ('-h', '--help'):
        _print_commands()
        sys.exit(0)
    elif cmdname not in cmds:
        _print_unknown_command(cmdname)
        sys.exit(2)
    del argv[1]
    cmd = cmds[cmdname]
    parser = argparse.ArgumentParser()
    parser.usage = "guniflask-manage {} {}".format(cmdname, cmd.syntax)
    parser.description = cmd.long_desc
    cmd.add_arguments(parser)
    try:
        args = parser.parse_args(args=argv[1:])
        cmd.process_arguments(args)
        cmd.run(args)
    except UsageError as e:
        print('Error: {}'.format(e), file=sys.stderr)
        sys.exit(2)
    else:
        if cmd.exitcode:
            sys.exit(cmd.exitcode)


if __name__ == '__main__':
    main()



# coding=utf-8

from datetime import datetime

__all__ = ['OAuth2AccessToken', 'OAuth2RefreshToken']


class OAuth2RefreshToken:
    """
    OAuth2 refresh token
    """

    def __init__(self, value: str, expiration: datetime = None):
        self.value = value
        self.expiration = expiration


class OAuth2AccessToken:
    """
    OAuth2 access token
    """

    BEARER_TYPE = 'Bearer'
    TOKEN_TYPE = 'token_type'
    ACCESS_TOKEN = 'access_token'
    REFRESH_TOKEN = 'refresh_token'
    EXPIRES_IN = 'expires_in'
    SCOPE = 'scope'

    def __init__(self, value, refresh_token: OAuth2RefreshToken = None, expiration: datetime = None, scope=None,
                 additional_info=None):
        self.value = value
        self.token_type = self.BEARER_TYPE.lower()
        self.refresh_token = refresh_token
        self.expiration = expiration
        self.scope = set()
        if scope is not None:
            self.scope.update(scope)
        self.additional_info = dict()
        if additional_info is not None:
            self.additional_info.update(additional_info)

    def __str__(self):
        return str(self.to_dict())

    @property
    def expires_in(self):
        if self.expiration is None:
            return 0
        return int(self.expiration.timestamp() - datetime.now().timestamp())

    def to_dict(self):
        res = {self.ACCESS_TOKEN: self.value, self.TOKEN_TYPE: self.token_type}
        if self.refresh_token is not None:
            res[self.REFRESH_TOKEN] = self.refresh_token.value
        if self.expiration is not None:
            res[self.EXPIRES_IN] = self.expires_in
        if self.scope is not None:
            res[self.SCOPE] = ' '.join(self.scope)
        if self.additional_info is not None:
            res.update(self.additional_info)
        return res



# coding=utf-8

import logging

from flask_sqlalchemy import SQLAlchemy

log = logging.getLogger(__name__)

db = SQLAlchemy()


def make_settings(app, settings):
    """
    This function is invoked before initializing app.
    """


def init_app(app, settings):
    """
    This function is invoked before running app.
    """
    db.init_app(app)



# coding=utf-8

from flask import Flask

from guniflask.context.bean_context import AnnotationConfigBeanContext
from guniflask.web.blueprint_post_processor import BlueprintPostProcessor
from guniflask.web.scheduling_config import WebAsyncConfiguration, WebSchedulingConfiguration

__all__ = ['AnnotationConfigWebApplicationContext']


class AnnotationConfigWebApplicationContext(AnnotationConfigBeanContext):
    def __init__(self, app: Flask):
        super().__init__()
        self.app = app

    def _post_process_bean_factory(self, bean_factory):
        super()._post_process_bean_factory(bean_factory)
        bean_factory.add_bean_post_processor(BlueprintPostProcessor(self.app))
        self._reader.register(WebAsyncConfiguration)
        self._reader.register(WebSchedulingConfiguration)



# coding=utf-8

__all__ = ['BeanPostProcessor']


class BeanPostProcessor:
    def post_process_before_instantiation(self, bean_type: type, bean_name: str):
        return None

    def post_process_before_initialization(self, bean, bean_name: str):
        return bean

    def post_process_after_initialization(self, bean, bean_name: str):
        return bean



# coding=utf-8

__all__ = ['UsageError', 'AbortedError', 'TemplateError']


class UsageError(Exception):
    """
    Usage error
    """

    def __init__(self, *args, print_help=False, **kwargs):
        self.print_help = print_help
        super().__init__(*args, **kwargs)


class AbortedError(Exception):
    """
    Aborted error
    """


class TemplateError(Exception):
    """
    Template error
    """



# coding=utf-8

from typing import List

from guniflask.beans.definition import BeanDefinition
from guniflask.beans.errors import NoSuchBeanDefinitionError, BeanDefinitionStoreError

__all__ = ['BeanDefinitionRegistry']


class BeanDefinitionRegistry:
    def __init__(self):
        self._bean_definition_map = {}
        self._allow_bean_definition_overriding = True

    def set_allow_bean_definition_overriding(self, allow_bean_definition_overriding: bool):
        self._allow_bean_definition_overriding = allow_bean_definition_overriding

    @property
    def is_allow_bean_definition_overriding(self) -> bool:
        return self._allow_bean_definition_overriding

    def register_bean_definition(self, bean_name: str, bean_definition: BeanDefinition):
        old_bean_definition = self._bean_definition_map.get(bean_name)
        if old_bean_definition is not None:
            if not self.is_allow_bean_definition_overriding:
                raise BeanDefinitionStoreError('A bean named "{}" is already bound'.format(bean_name))
        self._bean_definition_map[bean_name] = bean_definition

    def get_bean_definition(self, bean_name: str) -> BeanDefinition:
        bean_definition = self._bean_definition_map.get(bean_name)
        if bean_definition is None:
            raise NoSuchBeanDefinitionError(bean_name)
        return bean_definition

    def get_bean_definition_names(self) -> List[str]:
        return list(self._bean_definition_map.keys())

    def contains_bean_definition(self, bean_name: str) -> bool:
        return bean_name in self._bean_definition_map

    def remove_bean_definition(self, bean_name: str):
        if bean_name not in self._bean_definition_map:
            raise NoSuchBeanDefinitionError(bean_name)
        self._bean_definition_map.pop(bean_name)



# coding=utf-8

from flask import current_app, Flask
from functools import wraps, partial

from guniflask.context.annotation import bean, configuration
from guniflask.scheduling.config_constants import ASYNC_ANNOTATION_PROCESSOR, SCHEDULED_ANNOTATION_PROCESSOR
from guniflask.scheduling.annotation import AsyncRun, Scheduled
from guniflask.scheduling.async_config import AsyncConfiguration
from guniflask.scheduling.async_post_processor import AsyncPostProcessor
from guniflask.scheduling.scheduling_config import SchedulingConfiguration
from guniflask.scheduling.scheduling_post_processor import ScheduledPostProcessor

__all__ = ['WebAsyncConfiguration', 'WebAsyncPostProcessor',
           'WebSchedulingConfiguration', 'WebScheduledPostProcessor']


@configuration
class WebAsyncConfiguration(AsyncConfiguration):

    @bean(ASYNC_ANNOTATION_PROCESSOR)
    def async_annotation_processor(self) -> AsyncPostProcessor:
        return WebAsyncPostProcessor()


class WebAsyncPostProcessor(AsyncPostProcessor):

    def _process_async(self, async_run: AsyncRun, bean, method):
        app = current_app._get_current_object()
        wrapped_method = partial(run_with_app_context, method, app=app)
        wraps(method)(wrapped_method)
        super()._process_async(async_run, bean, wrapped_method)


@configuration
class WebSchedulingConfiguration(SchedulingConfiguration):

    @bean(SCHEDULED_ANNOTATION_PROCESSOR)
    def scheduled_annotation_processor(self) -> ScheduledPostProcessor:
        return WebScheduledPostProcessor()


class WebScheduledPostProcessor(ScheduledPostProcessor):

    def _process_scheduled(self, scheduled: Scheduled, method):
        app = current_app._get_current_object()
        wrapped_method = partial(run_with_app_context, method, app=app)
        wraps(method)(wrapped_method)
        super()._process_scheduled(scheduled, wrapped_method)


def run_with_app_context(func, app: Flask = None):
    if app is not None:
        with app.app_context():
            return func()
    return func()



# coding=utf-8

__all__ = ['ApplicationEvent',
           'ContextRefreshedEvent', 'ContextClosedEvent']


class ApplicationEvent:
    def __init__(self, source):
        self.source = source


class ContextRefreshedEvent(ApplicationEvent):
    pass


class ContextClosedEvent(ApplicationEvent):
    pass



# coding=utf-8

import os
import sys
from os.path import join, isfile, isdir
import re
import json

__all__ = ['set_default_env', 'infer_project_name', 'get_project_name_from_env']


def set_default_env():
    home_dir = os.environ.get('GUNIFLASK_HOME')
    if home_dir is None:
        home_dir = os.getcwd()
        os.environ['GUNIFLASK_HOME'] = home_dir
    if home_dir not in sys.path:
        sys.path.append(home_dir)
    if 'GUNIFLASK_PROJECT_NAME' not in os.environ:
        project_name = infer_project_name(home_dir)
        if project_name:
            os.environ['GUNIFLASK_PROJECT_NAME'] = project_name
    if 'GUNIFLASK_CONF_DIR' not in os.environ:
        os.environ['GUNIFLASK_CONF_DIR'] = join(home_dir, 'conf')
    if 'GUNIFLASK_LOG_DIR' not in os.environ:
        os.environ['GUNIFLASK_LOG_DIR'] = join(home_dir, '.log')
    if 'GUNIFLASK_PID_DIR' not in os.environ:
        os.environ['GUNIFLASK_PID_DIR'] = join(home_dir, '.pid')
    if 'GUNIFLASK_ID_STRING' not in os.environ:
        os.environ['GUNIFLASK_ID_STRING'] = os.getlogin()


project_name_regex = re.compile(r'[a-zA-Z\-]+')


def infer_project_name(home_dir):
    init_file = join(home_dir, '.guniflask-init.json')
    if isfile(init_file):
        try:
            with open(init_file, 'r') as f:
                data = json.load(f)
            project_name = data.get('project_name')
            if project_name and isdir(join(home_dir, project_name)):
                return project_name
        except Exception:
            pass
    candidates = []
    for d in os.listdir(home_dir):
        if project_name_regex.fullmatch(d) and isfile(join(home_dir, d, '__init__.py')) \
                and isfile(join(home_dir, d, 'app.py')):
            candidates.append(d)
    if len(candidates) == 0:
        raise RuntimeError('Cannot infer the project nam')


def get_project_name_from_env():
    return os.environ.get('GUNIFLASK_PROJECT_NAME')



# coding=utf-8

import os
from os.path import isfile, join, isdir
from importlib import import_module
from pkgutil import iter_modules


def walk_modules(path: str):
    mods = []
    mod = import_module(path)
    mods.append(mod)
    if hasattr(mod, '__path__'):
        for _, subpath, ispkg in iter_modules(mod.__path__):
            fullpath = path + '.' + subpath
            if ispkg:
                mods += walk_modules(fullpath)
            else:
                submod = import_module(fullpath)
                mods.append(submod)
    return mods


def walk_files(path: str):
    files = []
    if isdir(path):
        names = os.listdir(path)
        for name in names:
            files += walk_files(join(path, name))
    elif isfile(path):
        files.append(path)
    return files


def load_object(path):
    if isinstance(path, str):
        dot = path.rindex('.')
        module, name = path[:dot], path[dot + 1:]
        mod = import_module(module)
        return getattr(mod, name)
    return path



# coding=utf-8

from guniflask.security.request import OAuth2Request

__all__ = ['Authentication', 'OAuth2Authentication', 'UserAuthentication', 'PreAuthenticatedToken']


class Authentication:
    def __init__(self, authorities=None):
        self._authorities = set()
        if authorities is not None:
            self._authorities.update(authorities)
        self._details = None
        self._authenticated = False

    @property
    def principal(self):
        raise NotImplemented

    @property
    def credentials(self):
        raise NotImplemented

    def authenticate(self, value):
        self._authenticated = value

    @property
    def is_authenticated(self):
        return self._authenticated

    @property
    def authorities(self):
        return self._authorities

    @property
    def details(self):
        return self._details

    @details.setter
    def details(self, details):
        self._details = details


class UserAuthentication(Authentication):
    def __init__(self, principal, credentials=None, authorities=None):
        super().__init__(authorities=authorities)
        self._principal = principal
        self._credentials = credentials

    @property
    def principal(self):
        return self._principal

    @property
    def credentials(self):
        return self._credentials


class OAuth2Authentication(Authentication):
    def __init__(self, stored_request: OAuth2Request, user_authentication: Authentication = None):
        authorities = None
        if user_authentication is not None:
            authorities = user_authentication.authorities
        super().__init__(authorities=authorities)
        self._stored_request = stored_request
        self._user_authentication = user_authentication

    @property
    def principal(self):
        if self._user_authentication is None:
            return self._stored_request.client_id
        return self.user_authentication.principal

    @property
    def credentials(self):
        return None

    @property
    def stored_request(self):
        return self._stored_request

    @property
    def user_authentication(self):
        return self._user_authentication


class PreAuthenticatedToken(Authentication):
    def __init__(self, token):
        super().__init__()
        self._token = token

    @property
    def principal(self):
        return self._token

    @property
    def credentials(self):
        return None



# coding=utf-8

from functools import wraps

from werkzeug.exceptions import Unauthorized

from guniflask.security.user_details import current_user, AnonymousUser

__all__ = ['login_required', 'roles_required', 'authorities_required']


def login_required(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if isinstance(current_user._get_current_object(), AnonymousUser):
            raise Unauthorized
        return func(*args, **kwargs)

    return wrapper


def roles_required(*roles):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if not current_user.has_any_role(*roles):
                raise Unauthorized
            return func(*args, **kwargs)

        return wrapper

    return decorator


def authorities_required(*authorities):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if not current_user.has_any_authority(*authorities):
                raise Unauthorized
            return func(*args, **kwargs)

        return wrapper

    return decorator



# coding=utf-8

from flask import Flask, Blueprint as FlaskBlueprint

from guniflask.annotation.annotation_utils import AnnotationUtils
from guniflask.beans.post_processor import BeanPostProcessor
from guniflask.web.bind_annotation import Blueprint, Route

__all__ = ['BlueprintPostProcessor']


class BlueprintPostProcessor(BeanPostProcessor):
    def __init__(self, app: Flask):
        self.app = app

    def post_process_after_initialization(self, bean, bean_name: str):
        bean_type = bean.__class__
        annotation = AnnotationUtils.get_annotation(bean_type, Blueprint)
        if annotation is not None:
            options = annotation['options'] or {}
            b = FlaskBlueprint(bean_name, bean.__module__,
                               url_prefix=annotation['url_prefix'], **options)
            for m in dir(bean):
                method = getattr(bean, m)
                a = AnnotationUtils.get_annotation(method, Route)
                if a is not None:
                    rule_options = a['options'] or {}
                    b.add_url_rule(a['rule'], endpoint=method.__name__, view_func=method, **rule_options)
            self.app.register_blueprint(b)
        return bean



# coding=utf-8

from typing import Union, Collection

from guniflask.security.authentication import OAuth2Authentication
from guniflask.security.token import OAuth2AccessToken, OAuth2RefreshToken

__all__ = ['TokenStore', 'JwtTokenStore']


class TokenStore:
    """
    Persistence interface for OAuth2 tokens.
    """

    def read_authentication(self, access_token: Union[OAuth2AccessToken, str]) -> OAuth2AccessToken:
        raise NotImplemented

    def store_access_token(self, access_token: OAuth2AccessToken, authentication: OAuth2Authentication):
        raise NotImplemented

    def read_access_token(self, access_token_value: str) -> OAuth2AccessToken:
        raise NotImplemented

    def remove_access_token(self, access_token: OAuth2AccessToken):
        raise NotImplemented

    def store_refresh_token(self, refresh_token: str, authentication: OAuth2Authentication):
        raise NotImplemented

    def read_refresh_token(self, refresh_token_value: str) -> OAuth2RefreshToken:
        raise NotImplemented

    def read_authentication_for_refresh_token(self, refresh_token: OAuth2RefreshToken) -> OAuth2Authentication:
        raise NotImplemented

    def remove_refresh_token(self, refresh_token: OAuth2RefreshToken):
        raise NotImplemented

    def remove_access_token_using_refresh_token(self, refresh_token: OAuth2RefreshToken):
        raise NotImplemented

    def get_access_token(self, authentication: OAuth2Authentication) -> OAuth2AccessToken:
        raise NotImplemented

    def find_tokens_by_client_id_and_username(self, client_id: str, username: str) -> Collection:
        raise NotImplemented

    def find_tokens_by_client_id(self, client_id: str) -> Collection:
        raise NotImplemented


class JwtTokenStore(TokenStore):
    def __init__(self):
        self.jwt_token_converter = None

    def read_authentication(self, access_token):
        if isinstance(access_token, OAuth2AccessToken):
            access_token = access_token.value
        return self.jwt_token_converter.extract_authentication(self.jwt_token_converter.decode(access_token))

    def store_access_token(self, access_token, authentication):
        return None

    def read_access_token(self, access_token_value):
        access_token = self._convert_access_token(access_token_value)
        if self.jwt_token_converter.is_refresh_token(access_token):
            raise InvalidTokenError('Encoded token is a refresh token')
        return access_token

    def remove_access_token(self, access_token):
        return None

    def store_refresh_token(self, refresh_token, authentication):
        return None

    def read_refresh_token(self, refresh_token_value):
        # TODO
        pass

    def read_authentication_for_refresh_token(self, refresh_token):
        return self.read_authentication(refresh_token.value)

    def remove_refresh_token(self, refresh_token):
        return None

    def remove_access_token_using_refresh_token(self, refresh_token):
        return None

    def get_access_token(self, authentication):
        return None

    def find_tokens_by_client_id_and_username(self, client_id, username):
        return []

    def find_tokens_by_client_id(self, client_id):
        return []

    def _convert_access_token(self, access_token_value):
        return self.jwt_token_converter.extract_access_token(access_token_value,
                                                             self.jwt_token_converter.decode(access_token_value))



# coding=utf-8

import os
from os.path import join, dirname, exists
import multiprocessing

from gunicorn.config import KNOWN_SETTINGS
from gunicorn.app.base import Application

from guniflask.config.app_config import load_profile_config
from guniflask.utils.env import walk_files
from guniflask.app import create_app
from guniflask.cli.env import get_project_name_from_env

__all__ = ['GunicornApplication']


class GunicornApplication(Application):

    def __init__(self):
        self.options = self._make_options()
        super().__init__()

    def set_option(self, key, value):
        if key in self.cfg.settings:
            self.cfg.set(key, value)

    def load_config(self):
        for key, value in self.options.items():
            if key in self.cfg.settings and value is not None:
                self.cfg.set(key.lower(), value)

    def load(self):
        return create_app(get_project_name_from_env())

    def _make_options(self):
        pid_dir = os.environ['GUNIFLASK_PID_DIR']
        log_dir = os.environ['GUNIFLASK_LOG_DIR']
        id_string = os.environ['GUNIFLASK_ID_STRING']
        project_name = get_project_name_from_env()
        options = {
            'daemon': True,
            'workers': multiprocessing.cpu_count(),
            'worker_class': 'gevent',
            'pidfile': join(pid_dir, '{}-{}.pid'.format(project_name, id_string)),
            'accesslog': join(log_dir, '{}-{}.access.log'.format(project_name, id_string)),
            'errorlog': join(log_dir, '{}-{}.error.log'.format(project_name, id_string))
        }
        options.update(self._make_profile_options(os.environ.get('GUNIFLASK_ACTIVE_PROFILES')))
        # if debug
        if os.environ.get('GUNIFLASK_DEBUG'):
            options.update(self._make_debug_options())
        self._makedirs(options)
        # hook wrapper
        self._set_hook_wrapper(options)
        return options

    def _set_hook_wrapper(self, options):
        HookWrapper.from_config(options)

    def _make_profile_options(self, active_profiles):
        conf_dir = os.environ['GUNIFLASK_CONF_DIR']
        gc = load_profile_config(conf_dir, 'gunicorn', profiles=active_profiles)
        settings = {}
        snames = set([i.name for i in KNOWN_SETTINGS])
        for name in gc:
            if name in snames:
                settings[name] = gc[name]
        return settings

    @staticmethod
    def _make_debug_options():
        conf_dir = os.environ['GUNIFLASK_CONF_DIR']
        return {
            'accesslog': '-',
            'errorlog': '-',
            'loglevel': 'debug',
            'disable_redirect_access_to_syslog': True,
            'reload': True,
            'reload_extra_files': walk_files(conf_dir),
            'workers': 1,
            'daemon': False
        }

    @staticmethod
    def _makedirs(opts):
        for c in ['pidfile', 'accesslog', 'errorlog']:
            p = opts.get(c)
            if p:
                d = dirname(p)
                if d and not exists(d):
                    os.makedirs(d)


class HookWrapper:
    HOOKS = ['on_starting', 'on_reload', 'on_exit']

    def __init__(self, config, on_starting=None, on_reload=None, on_exit=None):
        self.config = config
        self._on_starting = on_starting
        self._on_reload = on_reload
        self._on_exit = on_exit

    @classmethod
    def from_config(cls, config):
        kw = {}
        for h in cls.HOOKS:
            kw[h] = config.get(h)
        wrapper = cls(config, **kw)
        for h in cls.HOOKS:
            config[h] = getattr(wrapper, h)
        return wrapper

    def on_starting(self, server):
        if self._on_starting is not None:
            self._on_starting(server)

    def on_reload(self, server):
        if self._on_reload is not None:
            self._on_reload(server)

    def on_exit(self, server):
        if self._on_exit is not None:
            self._on_exit(server)



# coding=utf-8

from typing import List, Type

__all__ = ['Annotation', 'AnnotationMetadata']


class Annotation:

    def __init__(self, **kwargs):
        self.attributes = {}
        self.attributes.update(kwargs)

    def __setitem__(self, key, value):
        self.attributes[key] = value

    def __getitem__(self, key):
        return self.attributes.get(key)


class AnnotationMetadata:
    key = '__annotation_metadata'

    def __init__(self, source):
        if not hasattr(source, '__name__'):
            raise ValueError('Cannot annotate to the object without "__name__" attribute')
        self._source = source
        self._annotations = {}

    @property
    def source(self):
        return self._source

    def add_annotation(self, annotation: Annotation):
        self._annotations[type(annotation)] = annotation

    def get_annotation(self, annotation_type: Type[Annotation]) -> Annotation:
        return self._annotations.get(annotation_type)

    @property
    def annotations(self) -> List[Annotation]:
        return list(self._annotations.values())

    def is_annotated(self, annotation_type: Type[Annotation]) -> bool:
        for a in self._annotations.values():
            if isinstance(a, annotation_type):
                return True
        return False



# coding=utf-8

__all__ = ['APPLICATION_EVENT_PUBLISHER',
           'CONFIGURATION_ANNOTATION_PROCESSOR',
           'CONFIGURATION_BEAN_NAME_GENERATOR']

APPLICATION_EVENT_PUBLISHER = '__application_event_publisher'

CONFIGURATION_ANNOTATION_PROCESSOR = '__configuration_annotation_processor'

CONFIGURATION_BEAN_NAME_GENERATOR = '__configuration_bean_name_generator'



# coding=utf-8

import os
from os.path import join, exists
from collections import defaultdict
from keyword import iskeyword
import re

import sqlalchemy
from sqlalchemy import ForeignKeyConstraint, UniqueConstraint, PrimaryKeyConstraint, CheckConstraint, ForeignKey
import inflect
from sqlalchemy.util import OrderedDict

from guniflask.model.database_dialect import supported_dialects
from guniflask.utils.string import string_camelcase, string_lowercase_underscore

__all__ = ['SqlToModelGenerator']

inflect_engine = inflect.engine()


class SqlToModelGenerator:
    def __init__(self, name, metadata, indent=4, bind=None):
        dialect_name = metadata.bind.dialect.name
        if dialect_name not in supported_dialects:
            raise ValueError('Unsupported dialect: {}'.format(dialect_name))

        self.dialect = supported_dialects[dialect_name]
        self.name = name
        self.metadata = metadata
        self.indent = ' ' * indent
        self.bind = bind

        many_to_many_tables = set()
        many_to_many_links = defaultdict(list)
        for table in metadata.tables.values():
            fk_constraints = [i for i in table.constraints if isinstance(i, ForeignKeyConstraint)]
            if len(fk_constraints) == 2 and all(col.foreign_keys for col in table.columns):
                many_to_many_tables.add(table.name)
                tablename = sorted(fk_constraints, key=get_constraint_sort_key)[0].elements[0].column.table.name
                many_to_many_links[tablename].append(table)

        self.models = {}
        for table in metadata.sorted_tables:
            if table.name in many_to_many_tables:
                continue
            self.models[table.name] = Model(table, many_to_many_links[table.name])

            # Add many-to-one relations
            for constraint in sorted(table.constraints, key=get_constraint_sort_key):
                if isinstance(constraint, ForeignKeyConstraint):
                    tablename = constraint.elements[0].column.table.name
                    self.models[tablename].add_many_to_one_relation(constraint)

    def render(self, path):
        if not exists(path):
            os.makedirs(path)
        model_modules = []
        for model in self.models.values():
            module_name = convert_to_valid_identifier(model.table.name)
            with open(join(path, module_name + '.py'), 'w', encoding='utf-8') as f:
                f.write('# coding=utf-8\n\n')
                f.write(self.render_imports(model))
                tables_content = self.render_secondary_tables(model)
                if tables_content:
                    f.write('\n')
                    f.write(tables_content)
                f.write('\n\n')
                f.write(self.render_model(model))
            model_modules.append({'module': module_name, 'class': model.class_name})
        with open(join(path, '__init__.py'), 'w', encoding='utf-8') as f:
            f.write('# coding=utf-8\n\n')
            for m in model_modules:
                f.write('from .{} import {}\n'.format(m['module'], m['class']))

    def render_imports(self, model):
        d = OrderedDict()
        for col in model.table.columns:
            if col.server_default:
                d.setdefault('sqlalchemy', ('text', '_text'))

        imports = ''
        for k, v in d.items():
            if isinstance(v, tuple):
                imports += 'from {} import {} as {}\n'.format(k, v[0], v[1])
            else:
                imports += 'from {} import {}\n'.format(k, v)
        if len(d) > 0:
            imports += '\n'
        imports += 'from {} import db\n'.format(self.name)
        return imports

    def render_model(self, model):
        header_str = 'class {}(db.Model):\n'.format(model.class_name)
        header_str += "{}__tablename__ = '{}'\n".format(self.indent, model.table.name)
        if self.bind:
            header_str += "{}__bind_key__ = '{}'\n".format(self.indent, self.bind)
        header_str += '\n'
        columns_str = ''
        for col in model.table.columns:
            attr = convert_to_valid_identifier(col.name)
            show_name = attr != col.name
            columns_str += '{}{} = {}\n'.format(self.indent, attr,
                                                self.render_column(col, show_name=show_name))
        relationships_str = ''
        for r in model.relationships:
            relationships_str += self.indent + self.render_relationship(r) + '\n'
        return header_str + columns_str + relationships_str

    def render_secondary_tables(self, model):
        return '\n'.join([self.render_table(r.association_table) for r in model.relationships
                          if isinstance(r, ManyToManyRelationship)])

    def render_table(self, table):
        columns_str = ',\n'.join(self.indent + self.render_column(col, show_name=True) for col in table.columns)
        tablename = convert_to_valid_identifier(table.name)
        return '{} = db.Table({!r},\n{}\n)\n'.format(tablename, table.name, columns_str)

    def render_column(self, column, show_name=False):
        is_sole_pk = column.primary_key and len(column.table.primary_key) == 1
        dedicated_fks = [c for c in column.foreign_keys if len(c.constraint.columns) == 1]
        is_unique = any(isinstance(c, UniqueConstraint) and set(c.columns) == {column}
                        for c in column.table.constraints)
        is_unique = is_unique or any(i.unique and set(i.columns) == {column}
                                     for i in column.table.indexes)
        has_index = any(set(i.columns) == {column} for i in column.table.indexes)
        server_default = None
        kwargs = []
        if column.key != column.name:
            kwargs.append('key')
        if column.primary_key:
            kwargs.append('primary_key')
        if column.autoincrement is True:
            kwargs.append('autoincrement')
        if not column.nullable and not is_sole_pk:
            kwargs.append('nullable')
        if is_unique:
            column.unique = True
            kwargs.append('unique')
        if has_index:
            column.index = True
            kwargs.append('index')
        if column.comment:
            kwargs.append('comment')
        if column.server_default:
            default_expr = self.get_compiled_expression(column.server_default.arg)
            if '\n' in default_expr:
                server_default = 'server_default=text("""\\\n{0}""")'.format(default_expr)
            else:
                default_expr = default_expr.replace('"', '\\"')
                server_default = 'server_default=_text("{0}")'.format(default_expr)
        extra_kwargs = self.get_extra_column_kwargs(column)
        return "db.Column({})".format(', '.join(([repr(column.name)] if show_name else []) +
                                                [self.render_column_type(column.type)] +
                                                [self.render_constraint(x) for x in dedicated_fks] +
                                                ['{}={!r}'.format(i, getattr(column, i)) for i in kwargs] +
                                                ([server_default] if server_default else []) +
                                                ['{}={}'.format(i, extra_kwargs[i]) for i in
                                                 sorted(extra_kwargs.keys())]))

    def get_extra_column_kwargs(self, column):
        kwargs = {}
        for p in supported_column_properties:
            if p.match_column(column):
                kwargs.update(p.kwargs)
        return kwargs

    def render_column_type(self, coltype):
        return 'db.{}'.format(self.dialect.convert_column_type(coltype))

    def render_constraint(self, constraint):
        def render_fk_options(*args):
            opts = [repr(i) for i in args]
            for attr in 'ondelete', 'onupdate':
                value = getattr(constraint, attr, None)
                if value:
                    opts.append('{}={!r}'.format(attr, value))
            return ', '.join(opts)

        if isinstance(constraint, ForeignKey):
            remote_column = '{}.{}'.format(constraint.column.table.name,
                                           constraint.column.name)
            return 'db.ForeignKey({})'.format(render_fk_options(remote_column))

    def render_relationship(self, relationship):
        kwargs_str = ', '.join([repr(table_name_to_class_name(relationship.target_tbl))] +
                               ['{}={}'.format(i, relationship.kwargs[i]) for i in
                                sorted(relationship.kwargs.keys())])
        return '{} = db.relationship({})'.format(relationship.preferred_name, kwargs_str)

    def get_compiled_expression(self, statement):
        return str(statement.compile(
            self.metadata.bind, compile_kwargs={"literal_binds": True}))


def convert_to_valid_identifier(name):
    name = string_lowercase_underscore(name)
    if name[0].isdigit() or iskeyword(name):
        name = '_' + name
    return name


def table_name_to_class_name(table_name):
    name = string_camelcase(table_name)
    if name[0].isdigit():
        name = '_' + name
    return name


def get_constraint_sort_key(constraint):
    if isinstance(constraint, CheckConstraint):
        return 'C{0}'.format(constraint.sqltext)
    return constraint.__class__.__name__[0] + repr(list(constraint.columns.keys()))


class Model:
    parent_name = 'db.Model'

    def __init__(self, table, association_tables):
        self.table = table
        self.class_name = table_name_to_class_name(table.name)
        self.relationships = []

        # Add many-to-many relationships
        for association_table in association_tables:
            fk_constraints = [c for c in association_table.constraints if isinstance(c, ForeignKeyConstraint)]
            fk_constraints.sort(key=get_constraint_sort_key)
            target_tbl = fk_constraints[1].elements[0].column.table.name
            relationship = ManyToManyRelationship(self.table.name, target_tbl, association_table)
            self.relationships.append(relationship)

    def add_many_to_one_relation(self, constraint):
        relationship = ManyToOneRelationship(self.table.name, constraint.table.name, constraint)
        self.relationships.append(relationship)


class Relationship:
    def __init__(self, source_tbl, target_tbl):
        self.source_tbl = source_tbl
        self.target_tbl = target_tbl
        self.kwargs = {}
        self.preferred_name = None


class ManyToOneRelationship(Relationship):
    def __init__(self, source_tbl, target_tbl, constraint):
        super().__init__(source_tbl, target_tbl)

        self.preferred_name = convert_to_valid_identifier(target_tbl)
        self.constraint = constraint

        # Add uselist=False to one-to-one relationships
        column_names = list(constraint.columns.keys())
        if any(isinstance(c, (PrimaryKeyConstraint, UniqueConstraint)) and
               set(col.name for col in c.columns) == set(column_names)
               for c in constraint.table.constraints):
            self.kwargs['uselist'] = 'False'

        if self.kwargs.get('uselist') is not False:
            self.preferred_name = inflect_engine.plural(self.preferred_name)
            self.kwargs['lazy'] = repr('select')
        else:
            self.kwargs['lazy'] = repr('joined')
        self.kwargs['backref'] = "db.backref({}, lazy='joined')".format(
            repr(convert_to_valid_identifier(source_tbl)))
        self.kwargs['cascade'] = repr('all, delete-orphan')


class ManyToManyRelationship(Relationship):
    def __init__(self, source_tbl, target_tbl, association_table):
        super().__init__(source_tbl, target_tbl)

        self.preferred_name = inflect_engine.plural(convert_to_valid_identifier(target_tbl))
        self.association_table = association_table

        self.kwargs['secondary'] = convert_to_valid_identifier(association_table.name)
        self.kwargs['lazy'] = repr('select')
        self.kwargs['backref'] = "db.backref({}, lazy='select')".format(
            repr(inflect_engine.plural(convert_to_valid_identifier(source_tbl))))


class ColumnProperty:
    def __init__(self):
        self.kwargs = {}

    def match_column(self, column):
        raise NotImplementedError


class CreatedTimeProperty(ColumnProperty):
    def __init__(self):
        super().__init__()
        self.kwargs['default'] = 'db.func.now()'
        self.reg = re.compile(r'^create[d]?_(time|at)$')

    def match_column(self, column):
        return isinstance(column.type, sqlalchemy.DateTime) and self.reg.match(column.name.lower()) is not None


class UpdatedTimeProperty(ColumnProperty):
    def __init__(self):
        super().__init__()
        self.kwargs['default'] = 'db.func.now()'
        self.kwargs['onupdate'] = 'db.func.now()'
        self.reg = re.compile(r'^update[d]?_(time|at)$')

    def match_column(self, column):
        return isinstance(column.type, sqlalchemy.DateTime) and self.reg.match(column.name.lower()) is not None


supported_column_properties = [CreatedTimeProperty(), UpdatedTimeProperty()]



# coding=utf-8

from typing import List, Type

from guniflask.annotation.core import Annotation, AnnotationMetadata

__all__ = ['AnnotationUtils']


class AnnotationUtils:

    @staticmethod
    def get_annotation_metadata(source) -> AnnotationMetadata:
        if hasattr(source, AnnotationMetadata.key):
            metadata = getattr(source, AnnotationMetadata.key)
            # annotations on class cannot be inherited
            if not isinstance(metadata.source, type) or metadata.source == source:
                return metadata

    @staticmethod
    def get_annotation(source, annotation_type: Type[Annotation]) -> Annotation:
        annotation_metadata = AnnotationUtils.get_annotation_metadata(source)
        if annotation_metadata is not None:
            return annotation_metadata.get_annotation(annotation_type)

    @staticmethod
    def get_annotations(source) -> List[Annotation]:
        annotation_metadata = AnnotationUtils.get_annotation_metadata(source)
        if annotation_metadata is not None:
            return annotation_metadata.annotations
        return []

    @staticmethod
    def add_annotation(source, annotation: Annotation):
        annotation_metadata = AnnotationUtils.get_annotation_metadata(source)
        if annotation_metadata is None:
            annotation_metadata = AnnotationMetadata(source)
            setattr(source, AnnotationMetadata.key, annotation_metadata)
        annotation_metadata.add_annotation(annotation)



# coding=utf-8

import datetime as dt

from guniflask.security.authentication import OAuth2Authentication
from guniflask.security.token import OAuth2AccessToken

__all__ = ['TokenServices']


class TokenServices:
    """
    Base implementation for token services using random UUID values for the access token and refresh token values.
    """

    def __init__(self):
        self.token_store = None
        self.token_converter = None
        self.client_details_service = None
        self.access_token_expires_in = dt.timedelta(days=1)
        self.refresh_token_expires_in = dt.timedelta(days=365)

    def create_access_token(self, authentication: OAuth2Authentication = None) -> OAuth2AccessToken:
        # TODO
        pass

    def refresh_access_token(self, refresh_token_value, token_request) -> OAuth2AccessToken:
        # TODO
        pass

    def get_access_token(self, authentication) -> OAuth2AccessToken:
        return self.token_store.get_access_token(authentication)

    def load_authentication(self, access_token_value) -> OAuth2Authentication:
        # TODO
        pass

    def read_access_token(self, access_token_value) -> OAuth2AccessToken:
        return self.token_store.read_access_token(access_token_value)

    def revoke_token(self, access_token_value):
        access_token = self.token_store.read_access_token(access_token_value)
        if access_token:
            if access_token.refresh_token:
                self.token_store.remove_refresh_token(access_token.refresh_token)
            self.token_store.remove_access_token(access_token)



# coding=utf-8

from .command import *
from .env import *
from .errors import *
from .step import *

__all__ = (command.__all__ +
           env.__all__ +
           errors.__all__ +
           step.__all__)



# coding=utf-8

__all__ = ['ClientDetails', 'ClientDetailsService']


class ClientDetails:
    """
    Client details for OAuth2
    """

    def __init__(self, client_id=None, client_secret=None, resources_ids=None, scopes=None, grant_types=None,
                 authorities=None, redirect_uris=None, auto_approve_scopes=None,
                 access_token_expires_in=None, refresh_token_expires_in=None,
                 additional_information=None):
        self.client_id = client_id
        self.client_secret = client_secret
        self.resource_ids = set()
        if resources_ids is not None:
            self.resource_ids.update(resources_ids)
        self.scopes = set()
        if scopes is not None:
            self.scopes.update(scopes)
        self.grant_types = set()
        if grant_types is not None:
            self.grant_types.update(grant_types)
        self.authorities = set()
        if authorities is not None:
            self.authorities.update(authorities)
        self.redirect_uris = set()
        if redirect_uris is not None:
            self.redirect_uris.update(redirect_uris)
        self.auto_approve_scopes = set()
        if auto_approve_scopes is not None:
            self.auto_approve_scopes.update(auto_approve_scopes)
        self.access_token_expires_in = access_token_expires_in
        self.refresh_token_expires_in = refresh_token_expires_in
        self.additional_information = {}
        if additional_information is not None:
            self.additional_information.update(additional_information)

    def is_scoped(self):
        return self.scopes is not None and len(self.scopes) > 0

    def is_auto_approve(self, scope):
        return self.auto_approve_scopes is not None and scope in self.auto_approve_scopes


class ClientDetailsService:
    """
    A service that provides the details about an OAuth2 client.
    """

    def __init__(self):
        self._client_details_store = {}

    def load_client_details_by_client_id(self, client_id):
        return self._client_details_store.get(client_id)

    def add_client_details(self, client_details):
        self._client_details_store[client_details.client_id] = client_details

    def remove_client_details(self, client_details):
        client_id = client_details.client_id
        if client_id in self._client_details_store:
            self._client_details_store.pop(client_id)



# coding=utf-8

__all__ = ['Command']


class Command:
    def __init__(self):
        self.exitcode = 0

    @property
    def name(self):
        return ""

    @property
    def syntax(self):
        return ""

    @property
    def short_desc(self):
        return ""

    @property
    def long_desc(self):
        return self.short_desc

    def add_arguments(self, parser):
        pass

    def process_arguments(self, args):
        pass

    def run(self, args):
        raise NotImplementedError



# coding=utf-8

from os.path import join

from flask import render_template

from guniflask.config.template import _template_folder
from guniflask.web import blueprint, get_route


@blueprint('/hello-world',
           template_folder=join(_template_folder, 'hello_world'))
class HelloWorld:
    @get_route('/')
    def home_page(self):
        """
        Home page
        """
        return render_template('index.html')



# coding=utf-8

from .annotation import *
from .async_config import *
from .async_executor import *
from .async_post_processor import *
from .config_constants import *
from .scheduling_config import *
from .scheduling_post_processor import *
from .task_scheduler import *

__all__ = (annotation.__all__ +
           async_config.__all__ +
           async_executor.__all__ +
           async_post_processor.__all__ +
           config_constants.__all__ +
           scheduling_config.__all__ +
           scheduling_post_processor.__all__ +
           task_scheduler.__all__)
